// file      : xsd/cxx/tree/stream-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/stream-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c,
              String const& name_ = "",
              String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          Boolean named (l.named ());

          if (named)
          {
            name = ename (l);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          os << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >& o," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "return o << static_cast< ::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > const& > (v);"
             << "}";
        }

      private:
        String
        item_type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
        String type_name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c,
               String const& name_ = "",
               String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          Boolean named (u.named ());

          if (named)
          {
            name = ename (u);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          String base (L"::xsd::cxx::tree::string< " + char_type + L" >");

          os << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >& o," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "return o << static_cast< " << base << " const& > (v);"
             << "}";
        }

      private:
        String name;
        String type_name;
      };


      struct Enumerator : Traversal::Enumerator, protected virtual Context
      {
        Enumerator (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String const& name (ename (e));

          os << "case " << scope << "::" << name << ":"
             << "{"
             << "o << " << L << "\"" << e.name () << "\";"
             << "break;"
             << "}";
        }

      private:
        String scope;
      };

      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     String const& name_ = "",
                     String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          Boolean named (e.named ());

          if (named)
          {
            name = ename (e);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          os << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >& o," << endl
             << type_name << "::_xsd_" << name << " i)"
             << "{"
             << "switch (i)"
             << "{";

          {
            enter_scope (name);

            Traversal::Names names_enumerator;
            Enumerator enumerator (*this, type_name);

            names_enumerator >> enumerator;

            names (e, names_enumerator);

            leave_scope ();
          }


          os << "}" // switch
             << "return o;"
             << "}";

          os << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >& o," << endl
             << (named ? "class " : "") << type_name << " i)"
             << "{"
             << "return o << " <<
            type_name << "::_xsd_" << name << " (i);"
             << "}";
        }

      private:
        String name;
        String type_name;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            // I need to build scope information with the same names
            // as in the header so that I get member escaping right.
            // Actual names will be from the public interface.
            //

            String const& name (ename (m));
            String type_name (scope
                              + L"::" + name
                              + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }

      private:
        String scope;
      };


      struct Element : Traversal::Element, protected virtual Context
      {
        Element (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (skip (e)) return;

          String const& name (ename (e));

          if (max (e) != 1)
          {
            // sequence
            //
            os << "for (" << scope << "::" << name << "::" <<
              unclash (name, "const_iterator") << endl
               << "b (i." << name << " ().begin ()), " <<
              "e (i." << name << " ().end ());" << endl
               << "b != e; ++b)"
               << "{"
               << "o << ::std::endl << " << L << "\"" << e.name () <<
              ": \" << *b;"
               << "}";
          }
          else if (min (e) == 0)
          {
            // optional
            //

            os << "if (i." << name << " ())"
               << "{"
               << "o << ::std::endl << " << L << "\"" << e.name () <<
              ": \" << *i." << name << " ();"
               << "}";
          }
          else
          {
            // one
            //
            os << "o << ::std::endl << " << L << "\"" << e.name () <<
              ": \" << i." << name << " ();";
          }
        }

      private:
        String scope;
      };

      struct Attribute : Traversal::Attribute, protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));

          if (a.optional () && !a.default_ ())
          {
            os << "if (i." << name << " ())"
               << "{"
               << "o << ::std::endl << " << L << "\"" << a.name () <<
              ": \" << *i." << name << " ();"
               << "}";
          }
          else
          {
            os << "o << ::std::endl << " << L << "\"" << a.name () <<
              ": \" << i." << name << " ();";
          }
        }
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c,
                 String const& name_ = "",
                 String const& type_name_ = L"")
            : Context (c),
              name (name_),
              type_name (type_name_),
              base_ (c)
        {
          inherits_ >> base_;
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean named (c.named ());

          if (named)
          {
            name = ename (c);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // Go after anonymous types first.
          //
          {
            enter_scope (name);

            Traversal::Names names_member;
            Member<Complex> member (*this, type_name);

            names_member >> member;

            names (c, names_member);

            leave_scope ();
          }

          //
          //
          Boolean has_body (has<Traversal::Member> (c) || c.inherits_p ());

          os << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >& o," << endl
             << (named ? "class " : "") << type_name << " const&" <<
            (has_body ? " i" : "") << ")"
             << "{";

          inherits (c,
                    inherits_,
                    &Complex::inherits_pre,
                    &Complex::inherits_post,
                    &Complex::inherits_none);

          {
            enter_scope (name);

            Traversal::Names names_member;
            Element element (*this, type_name);
            Attribute attribute (*this);

            names_member >> element;
            names_member >> attribute;

            names (c, names_member);

            leave_scope ();
          }

          os << "return o;"
             << "}";
        }

        virtual Void
        inherits_pre (Type& t)
        {
          os << "o << static_cast< ";
        }

        virtual Void
        inherits_post (Type& t)
        {
          os << " const& > (i);";
        }

        virtual Void
        inherits_none (Type& t)
        {
        }

      private:
        String name;
        String type_name;

        Traversal::Inherits inherits_;
        BaseTypeName base_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
          : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String const& name (ename (m));
            String type_name (name + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };
    }

    Void
    generate_stream_source (Context& ctx,
                            SemanticGraph::Schema& tu,
                            UnsignedLong first,
                            UnsignedLong last)
    {
      String c (ctx.char_type);

      ctx.os << "#include <iostream> // std::basic_ostream" << endl
             << endl;


      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx, false, first, last);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;


      schema.dispatch (tu);
    }
  }
}
